"""Basic functions for searching for keywords in an ascii control file such as a nexus deck.

Ascii file must already have been opened for reading before calling any of these functions.
"""

version = '4th September 2020'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

# defs:
# def substring(shorter, longer):
# def end_of_file(ascii_file):
# def find_keyword(ascii_file, keyword, max_lines = None):
# def skip_blank_lines_and_comments(ascii_file, comment_intro = '!'):
# def split_trailing_comment(line, comment_char = '!'):
# def strip_trailing_comment(line, comment_char = '!'):
# def find_keyword_without_passing(ascii_file, keyword, no_pass_keyword):
# def find_keyword_with_copy(ascii_file_in, keyword, ascii_file_out):
# def find_keyword_pair(ascii_file, primary_keyword, secondary_keyword):
# def find_number(ascii_file):
# def specific_keyword_next(ascii_file, keyword, skip_blank_lines = True, comment_char = '!'):
# def number_next(ascii_file, skip_blank_lines = True, comment_char = '!'):
# def guess_comment_char(ascii_file):


def substring(shorter, longer):
    """Returns True if the first argument is a substring of the second."""
    try:
        longer.index(shorter)
        return True
    except Exception:
        return False


def end_of_file(ascii_file):
    """Returns True if the end of the file has been reached."""
    file_pos = ascii_file.tell()
    line = ascii_file.readline()
    if len(line) == 0:
        return True  # end of file
    ascii_file.seek(file_pos)
    return False


def find_keyword(ascii_file, keyword, max_lines = None):
    """Looks for line starting with given keyword; file pointer is left at start of that line."""
    start_pos = ascii_file.tell()
    while True:
        if max_lines is not None:
            if max_lines <= 0:
                ascii_file.seek(start_pos)
                return False
            max_lines -= 1
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            ascii_file.seek(start_pos)
            return False  # end of file
        words = line.split()
        if len(words) > 0 and words[0].upper() == keyword.upper():
            ascii_file.seek(file_pos)
            return True


def skip_blank_lines_and_comments(ascii_file, comment_char = '!', skip_c_space = True):
    """Skips any lines containing only white space or comment."""
    while True:
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            return  # end of file
        words = line.split()
        if len(words) == 0:
            continue
        if words[0][0] == comment_char:
            continue
        if skip_c_space and len(words[0]) == 1 and (words[0][0] == 'C' or words[0][0] == 'c'):
            continue
        ascii_file.seek(file_pos)
        return


def skip_comments(ascii_file, comment_char = '!', skip_c_space = True):
    """Skips any lines containing only a comment."""
    while True:
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            return  # end of file
        words = line.split()
        if len(words) > 0:  # not a blank line
            if words[0][0] == comment_char:
                continue
            if skip_c_space and len(words[0]) == 1 and (words[0][0] == 'C' or words[0][0] == 'c'):
                continue
        ascii_file.seek(file_pos)
        return


def split_trailing_comment(line, comment_char = '!'):
    """Returns a pair of strings: (line stripped of trailing comment, trailing comment)."""
    # also removes trailing newline
    comment = ''
    local_line = line
    while len(local_line) > 0 and local_line[-1] in ['\n', '\r']:
        local_line = local_line[:-1]
    try:
        pling = local_line.index(comment_char)
        if pling < len(local_line) - 2:
            comment = local_line[pling + 1:]
        return (local_line[:pling], comment)
    except Exception:
        return (local_line, '')


def strip_trailing_comment(line, comment_char = '!'):
    """Returns a copy of line with any trailing comment removed."""
    (result, comment) = split_trailing_comment(line, comment_char = comment_char)
    return result


def find_keyword_without_passing(ascii_file, keyword, no_pass_keyword):
    """Looks for line starting with keyword, but without passing line starting with no_pass_keyword."""
    while True:
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            return False  # end of file
        words = line.split()
        if len(words) > 0:
            if words[0].upper() == keyword.upper():
                ascii_file.seek(file_pos)
                return True
            if words[0].upper() == no_pass_keyword.upper():
                ascii_file.seek(file_pos)
                return False


def find_keyword_with_copy(ascii_file_in, keyword, ascii_file_out):
    """Looks for line starting with given keyword, copying lines in the meantime."""
    while True:
        file_pos = ascii_file_in.tell()
        line = ascii_file_in.readline()
        if len(line) == 0:
            return False  # end of file
        words = line.split()
        if len(words) > 0 and words[0].upper() == keyword.upper():
            ascii_file_in.seek(file_pos)
            return True
        else:
            ascii_file_out.write(line)


def find_keyword_pair(ascii_file, primary_keyword, secondary_keyword):
    """Looks for line starting with a given pair of keywords."""
    while True:
        if not find_keyword(ascii_file, primary_keyword):
            return False
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            return False  # end of file
        words = line.split()
        if len(words) >= 2 and words[1].upper() == secondary_keyword.upper():
            ascii_file.seek(file_pos)
            return True


def find_number(ascii_file):
    """Looks for line starting with any number."""
    while True:
        file_pos = ascii_file.tell()
        line = ascii_file.readline()
        if len(line) == 0:
            return False  # end of file
        words = line.split()
        if len(words) > 0:
            try:
                float(words[0])
            except Exception:  # todo: should only catch a particular exception (type conversion)
                continue
            else:
                ascii_file.seek(file_pos)
                return True


def specific_keyword_next(ascii_file, keyword, skip_blank_lines = True, comment_char = '!'):
    """Returns True if next token in file is the specified keyword."""
    file_pos = ascii_file.tell()  # will restore file pos to original, which may be before blank lines
    while True:
        if (skip_blank_lines):
            file_pos = ascii_file.tell()  # will advance file pos to next non-blank line
        line = ascii_file.readline()
        if len(line) == 0:  # end of file
            ascii_file.seek(file_pos)
            return False
        words = line.split()
        if len(words) == 0 or words[0][0] == comment_char:
            continue  # blank line or comment
        ascii_file.seek(file_pos)  # restore file position whether or not keyword match
        if words[0].upper() == keyword.upper():
            return True
        return False


def number_next(ascii_file, skip_blank_lines = True, comment_char = '!'):
    """Returns True if next token in file is a number."""
    file_pos = ascii_file.tell()
    while True:
        if (skip_blank_lines):
            file_pos = ascii_file.tell()  # will advance file pos to next non-blank line
        line = ascii_file.readline()
        if len(line) == 0:  # end of file
            ascii_file.seek(file_pos)
            return False
        words = line.split()
        if len(words) == 0 or words[0][0] == comment_char:
            continue  # blank line or comment
        ascii_file.seek(file_pos)  # restore file position whether or not number found
        try:
            float(words[0])
        except Exception:  # todo: should only catch a particular exception (type conversion)
            return False
        else:
            return True


def blank_line(ascii_file):
    """Returns True if the next line contains only white space; False otherwise (including comments)."""

    file_pos = ascii_file.tell()
    line = ascii_file.readline()
    ascii_file.seek(file_pos)
    if len(line) == 0:
        return True  # end of file
    words = line.split()
    return len(words) == 0


def guess_comment_char(ascii_file):
    """Returns a string (usually one character) being the guess as to the comment character, or None."""
    file_pos = ascii_file.tell()  # will restore file pos to original
    max_lines = 10
    ch = None
    while max_lines:
        line = ascii_file.readline()
        max_lines -= 1
        if not line:
            break
        words = line.split()
        if len(words) and words[0] in ['!', 'C', '#', '--']:
            ch = words[0]
            break
    ascii_file.seek(file_pos)
    return ch


# end of keyword_files module



"""write_hdf5.py: Class to write a resqml hdf5 file and functions for copying hdf5 data."""

version = '26th August 2021'

# Nexus is a registered trademark of the Halliburton Company

# write_hdf5.py
# approach is to register the datasets (arrays) to be written; then write everything in a separate, single function call

import logging

log = logging.getLogger(__name__)

import os

import h5py
import numpy as np

import resqpy.olio.uuid as bu

resqml_path_head = '/RESQML/'  # note: latest fesapi code uses RESQML20
write_bool_as_int8 = True  # Nexus read fails if bool used as hdf5 element dtype; also better for NullValue handling
write_int_as_int32 = True  # only applies if registered dtype is None


class H5Register():
    """Class for registering arrays and then writing to an hdf5 file."""

    def __init__(self, model):
        """Create a new, empty register of arrays to be written to an hdf5 file."""

        self.dataset_dict = {}  # dictionary mapping from (object_uuid, group_tail) to (numpy_array, dtype)
        self.hdf5_path_dict = {}  # dictionary optionally mapping from (object_uuid, group_tail) to hdf5 internal path
        self.model = model

    def register_dataset(self, object_uuid, group_tail, a, dtype = None, hdf5_path = None, copy = False):
        """Register an array to be included as a dataset in the hdf5 file.

        arguments:
           object_uuid (uuid.UUID): the uuid of the object (part) that this array is for
           group_tail (string): the remainder of the hdf5 internal path (following RESQML and
              uuid elements)
           a (numpy array): the dataset (array) to be registered for writing
           dtype (type or string): the required type of the individual elements within the dataset
           hdf5_path (string, optional): if present, a full hdf5 internal path to use instead of
              the default generated from the uuid
           copy (boolean, default False): if True, a copy of the array will be made at the time of
              registering, otherwise changes made to the array before the write() method is called
              are likely to be in the data that is written

        returns:
           None

        notes:
           several arrays might belong to the same object;
           if a dtype is given and necessitates a conversion of the array data, the behaviour will
           be as if the copy argument is True regardless of its setting
        """

        #     print('registering dataset with uuid ' + str(object_uuid) + ' and group tail ' + group_tail)
        assert (len(group_tail) > 0)
        assert a is not None
        assert isinstance(a, np.ndarray)
        if dtype is not None:
            a = a.astype(dtype, copy = copy)
        elif copy:
            a = a.copy()
        if group_tail[0] == '/':
            group_tail = group_tail[1:]
        if group_tail[-1] == '/':
            group_tail = group_tail[:-1]
        if (object_uuid, group_tail) in self.dataset_dict.keys():
            pass  # todo: warn of re-registration?
        self.dataset_dict[(object_uuid, group_tail)] = (a, dtype)
        if hdf5_path:
            self.hdf5_path_dict[(object_uuid, group_tail)] = hdf5_path

    def write_fp(self, fp):
        """Write or append to an hdf5 file, writing the pre-registered datasets (arrays).

        arguments:
           fp: an already open h5py._hl.files.File object

        returns:
           None

        note:
           the file handle fp must have been opened with mode 'w' or 'a'
        """

        # note: in resqml, an established hdf5 file has a uuid and should therefore be immutable
        #       this function allows appending to any hdf5 file; calling code should set a new uuid when needed
        assert (fp is not None)
        for (object_uuid, group_tail) in self.dataset_dict.keys():
            if (object_uuid, group_tail) in self.hdf5_path_dict.keys():
                hdf5_path = self.hdf5_path_dict[(object_uuid, group_tail)]
            else:
                hdf5_path = resqml_path_head + str(object_uuid) + '/' + group_tail
            (a, dtype) = self.dataset_dict[(object_uuid, group_tail)]
            if dtype is None:
                dtype = a.dtype
                if write_int_as_int32 and str(dtype) == 'int64':
                    dtype = 'int32'
            if write_bool_as_int8 and str(dtype).lower().startswith('bool'):
                dtype = 'int8'
            log.debug('Writing hdf5 dataset ' + hdf5_path + ' of size ' + str(a.size) + ' type ' + str(dtype))
            fp.create_dataset(hdf5_path, data = a, dtype = dtype)

    def write(self, file = None, mode = 'w', release_after = True):
        """Create or append to an hdf5 file, writing the pre-registered datasets (arrays).

        arguments:
           file: either a string being the file path, or an already open h5py._hl.files.File object;
              if None (recommended), the file is opened through the model object's hdf5 management
              functions
           mode (string, default 'w'): the mode to open the file in; only relevant if file is a path;
              must be 'w' or 'a' for (over)write or append

        returns:
           None
        """

        # note: in resqml, an established hdf5 file has a uuid and should therefore be immutable
        #       this function allows appending to any hdf5 file;
        #       strictly, calling code should set a new uuid when needed, in practice not essential
        if len(self.dataset_dict) == 0:
            return
        if file is None:
            file = self.model.h5_access(mode = mode)
        elif isinstance(file, str):
            log.debug(f'writing to hdf5 file: {file}')
            file = self.model.h5_access(mode = mode, file_path = file)
        if mode == 'a' and isinstance(file, str) and not os.path.exists(file):
            mode = 'w'
        assert isinstance(file, h5py._hl.files.File)
        self.write_fp(file)
        if release_after:
            self.model.h5_release()


def copy_h5(file_in, file_out, uuid_inclusion_list = None, uuid_exclusion_list = None, mode = 'w'):
    """Create a copy of an hdf5, optionally including or excluding arrays with specified uuids.

    arguments:
       file_in (string): path of existing hdf5 file to be duplicated
       file_out (string): path of output hdf5 file to be created or appended to (see mode)
       uuid_inclusion_list (list of uuid.UUID, optional): if present, the uuids to be included
          in the output file
       uuid_exclusion_list (list of uuid.UUID, optional): if present, the uuids to be excluded
          from the output file
       mode (string, default 'w'): mode to open output file with; must be 'w' or 'a' for
          (over)write or append respectively

    returns:
       number of hdf5 groups (uuid's) copied

    notes:
       at most one of uuid_inclusion_list and uuid_exclusion_list should be passed;
       if neither are passed, all the datasets (arrays) in the input file are copied to the
       output file
    """

    #  note: if both inclusion and exclusion lists are present, exclusion list is ignored
    assert file_out != file_in, 'identical input and output files specified for hdf5 copy'
    assert uuid_inclusion_list is None or uuid_exclusion_list is None,  \
       'inclusion and exclusion lists both specified for hdf5 copy; at most one allowed'
    checking_uuid = uuid_inclusion_list is not None or uuid_exclusion_list is not None
    assert mode in ['w', 'a']
    copy_count = 0
    with h5py.File(file_out, mode) as fp_out:
        assert fp_out is not None, 'failed to open output hdf5 file: ' + file_out
        with h5py.File(file_in, 'r') as fp_in:
            assert fp_in is not None, 'failed to open input hdf5 file: ' + file_in
            main_group_in = fp_in['RESQML']
            assert main_group_in is not None, 'failed to find RESQML group in hdf5 file: ' + file_in
            if mode == 'w':
                main_group_out = fp_out.create_group('RESQML')
            elif mode == 'a':
                try:
                    main_group_out = fp_out['RESQML']
                except Exception:
                    main_group_out = fp_out.create_group('RESQML')
            else:
                main_group_out = fp_out['RESQML']
            for group in main_group_in:
                if checking_uuid:
                    uuid = bu.uuid_from_string(group)
                    if uuid_inclusion_list is not None:
                        if uuid not in uuid_inclusion_list:
                            if uuid is None:
                                log.warning('RESQML group name in hdf5 file does not start with a uuid, skipping: ' +
                                            str(group))
                            continue
                    else:  # uuid_exclusion_list is not None
                        if uuid in uuid_exclusion_list:
                            continue
                        if uuid is None:  # will still be copied
                            log.warning('RESQML group name in hdf5 file does not start with a uuid: ' + str(group))
                if group in main_group_out:
                    log.warning('not copying hdf5 data due to pre-existence for: ' + str(group))
                    continue
                # log.debug('copying hdf5 data for uuid: ' + group)
                main_group_in.copy(group,
                                   main_group_out,
                                   expand_soft = True,
                                   expand_external = True,
                                   expand_refs = True)
                copy_count += 1
    return copy_count


def copy_h5_path_list(file_in, file_out, hdf5_path_list, mode = 'w'):
    """Create a copy of some hdf5 datasets (or groups), identified as a list of hdf5 internal paths.

    arguments:
       file_in (string): path of existing hdf5 file to be copied from
       file_out (string): path of output hdf5 file to be created or appended to (see mode)
       hdf5_path_list (list of string): the hdf5 internal paths of the datasets (or groups) to be copied
       mode (string, default 'w'): mode to open output file with; must be 'w' or 'a' for
          (over)write or append respectively

    returns:
       number of hdf5 datasets (or groups) copied
    """

    #  note: if both inclusion and exclusion lists are present, exclusion list is ignored
    assert file_out != file_in, 'identical input and output files specified for hdf5 copy'
    assert hdf5_path_list is not None
    assert mode in ['w', 'a']
    copy_count = 0
    with h5py.File(file_out, mode) as fp_out:
        assert fp_out is not None, f'failed to open output hdf5 file: {file_out}'
        with h5py.File(file_in, 'r') as fp_in:
            assert fp_in is not None, f'failed to open input hdf5 file: {file_in}'
            for path in hdf5_path_list:
                if path in fp_out:
                    log.warning(f'not copying hdf5 data due to pre-existence for: {path}')
                    continue
                assert path in fp_in, f'internal path {path} not found in hdf5 file {file_in}'
                log.debug(f'copying hdf5 data for: {path}')
                build = ''
                group_list = list(path.split(sep = '/'))
                assert len(group_list) > 1, f'no hdf5 group(s) in internal path {path}'
                for w in group_list[:-1]:
                    if w:
                        build += '/' + w
                        if build not in fp_out:
                            fp_out.create_group(build)
                build += '/' + group_list[-1]
                fp_out.create_dataset(build, data = fp_in[path])
                #            fp_in.copy(path, fp_out[path], expand_soft = True, expand_external = True, expand_refs = True)
                copy_count += 1
    return copy_count


def change_uuid(file, old_uuid, new_uuid):
    """Changes hdf5 internal path (group name) for part, switching from old to new uuid.

    notes:
       this is low level functionality not usually called directly;
       the function assumes that hdf5 internal path names conform to the format that resqpy uses
       when writing data, namely /RESQML/uuid/tail...
    """

    assert file, 'hdf5 file name missing'
    assert old_uuid is not None and new_uuid is not None, 'missing uuid'

    def change_uuid_fp(fp, old_uuid, new_uuid):
        main_group = fp[resqml_path_head.strip('/')]
        old_group = main_group[str(old_uuid)]
        main_group[str(new_uuid)] = old_group
        del main_group[str(old_uuid)]

    if isinstance(file, h5py._hl.files.File):
        change_uuid_fp(file, old_uuid, new_uuid)
    else:
        assert isinstance(file, str)
        with h5py.File(file, 'r+') as fp:
            change_uuid_fp(fp, old_uuid, new_uuid)



"""High level refined grid function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.fine_coarse as fc
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
import resqpy.rq_import as rqi

from resqpy.derived_model._common import _write_grid


def refined_grid(epc_file,
                 source_grid,
                 fine_coarse,
                 inherit_properties = False,
                 inherit_realization = None,
                 inherit_all_realizations = False,
                 source_grid_uuid = None,
                 set_parent_window = None,
                 infill_missing_geometry = True,
                 new_grid_title = None,
                 new_epc_file = None):
    """Generates a refined version of the source grid, optionally inheriting properties.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid unless source_grid_uuid is specified to identify the grid
       fine_coarse (resqpy.olio.fine_coarse.FineCoarse object): the mapping between cells in the fine (output) and
          coarse (source) grids
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid, with values resampled in the simplest way onto the finer grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       source_grid_uuid (uuid.UUID, optional): the uuid of the source grid – an alternative to the source_grid argument
          as a way of identifying the grid
       set_parent_window (boolean or str, optional): if True or 'parent', the refined grid has its parent window attribute
          set; if False, the parent window is not set; if None, the default will be True if new_epc_file is None or False
          otherwise; if 'grandparent' then an intervening parent window with no refinement or coarsening will be skipped
          and its box used in the parent window for the new grid, relating directly to the original grid
       infill_missing_geometry (boolean, default True): if True, an attempt is made to generate grid geometry in the
          source grid wherever it is undefined; if False, any undefined geometry will result in an assertion failure
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the refined grid (& crs)

    returns:
       new grid object being the refined grid; the epc and hdf5 files are written to as an intentional side effect

    notes:
       this function refines an entire grid; to refine a local area of a grid, first use the extract_box function
       and then use this function on the extracted grid; in such a case, using a value of 'grandparent' for the
       set_parent_window argument will relate the refined grid back to the original;
       if geometry infilling takes place, cached geometry and mask arrays within the source grid object will be
       modified as a side-effect of the function (but not written to hdf5 or changed in xml)
    """

    epc_file, model, model_in, source_grid =  \
        _establish_models_and_source_grid(epc_file, new_epc_file, source_grid, source_grid_uuid)

    assert fine_coarse is not None and isinstance(fine_coarse, fc.FineCoarse)
    if set_parent_window is None:
        set_parent_window = (new_epc_file is None)

    if infill_missing_geometry and (not source_grid.geometry_defined_for_all_cells() or
                                    not source_grid.geometry_defined_for_all_pillars()):
        log.debug('attempting infill of geometry missing in source grid')
        source_grid.set_geometry_is_defined(treat_as_nan = None,
                                            treat_dots_as_nan = True,
                                            complete_partial_pillars = True,
                                            nullify_partial_pillars = False,
                                            complete_all = True)
    assert source_grid.geometry_defined_for_all_pillars(), 'refinement requires geometry to be defined for all pillars'
    assert source_grid.geometry_defined_for_all_cells(), 'refinement requires geometry to be defined for all cells'

    assert tuple(fine_coarse.coarse_extent_kji) == tuple(source_grid.extent_kji),  \
           'fine_coarse mapping coarse extent does not match that of source grid'
    fine_coarse.assert_valid()

    source_grid.cache_all_geometry_arrays()
    if source_grid.has_split_coordinate_lines:
        source_grid.create_column_pillar_mapping()

    if model is not model_in:
        crs_part = model_in.part_for_uuid(source_grid.crs_uuid)
        assert crs_part is not None
        model.copy_part_from_other_model(model_in, crs_part)

    # todo: set nan-abled numpy operations?

    if source_grid.has_split_coordinate_lines:

        grid = _refined_faulted_grid(model, source_grid, fine_coarse)

    else:

        grid = _refined_unfaulted_grid(model, source_grid, fine_coarse)

    # todo: option of re-draping interpolated pillars to surface

    collection = None
    if inherit_properties:
        collection = _inherit_properties(source_grid, grid, fine_coarse, inherit_realization, inherit_all_realizations)

    if set_parent_window:
        _set_parent_window(set_parent_window, source_grid, grid, fine_coarse)

    # write grid
    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'grid refined from ' + str(rqet.citation_title_for_node(source_grid.root))

    model.h5_release()
    if model is not model_in:
        model_in.h5_release()

    if new_epc_file:
        _write_grid(new_epc_file, grid, property_collection = collection, grid_title = new_grid_title, mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        _write_grid(epc_file,
                    grid,
                    ext_uuid = ext_uuid,
                    property_collection = collection,
                    grid_title = new_grid_title,
                    mode = 'a')

    return grid


def _refined_faulted_grid(model, source_grid, fine_coarse):

    source_grid.corner_points(cache_cp_array = True)
    fnk, fnj, fni = fine_coarse.fine_extent_kji
    fine_cp = np.empty((fnk, fnj, fni, 2, 2, 2, 3))
    for ck0 in range(source_grid.nk):
        fine_k_base = fine_coarse.fine_base_for_coarse_axial(0, ck0)
        k_ratio = fine_coarse.ratio(0, ck0)
        k_interp = np.ones((k_ratio + 1,))
        k_interp[:-1] = fine_coarse.interpolation(0, ck0)
        for cj0 in range(source_grid.nj):
            fine_j_base = fine_coarse.fine_base_for_coarse_axial(1, cj0)
            j_ratio = fine_coarse.ratio(1, cj0)
            j_interp = np.ones((j_ratio + 1,))
            j_interp[:-1] = fine_coarse.interpolation(1, cj0)
            for ci0 in range(source_grid.ni):
                fine_i_base = fine_coarse.fine_base_for_coarse_axial(2, ci0)
                i_ratio = fine_coarse.ratio(2, ci0)
                i_interpolation = fine_coarse.interpolation(2, ci0)
                i_interp = np.ones((i_ratio + 1,))
                i_interp[:-1] = fine_coarse.interpolation(2, ci0)

                shared_fine_points = source_grid.interpolated_points((ck0, cj0, ci0), (k_interp, j_interp, i_interp))

                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 0, 0, 0] =  \
                   shared_fine_points[:-1, :-1, :-1]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 0, 0, 1] =  \
                   shared_fine_points[:-1, :-1, 1:]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 0, 1, 0] =  \
                   shared_fine_points[:-1, 1:, :-1]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 0, 1, 1] =  \
                   shared_fine_points[:-1, 1:, 1:]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 1, 0, 0] =  \
                   shared_fine_points[1:, :-1, :-1]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 1, 0, 1] =  \
                   shared_fine_points[1:, :-1, 1:]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 1, 1, 0] =  \
                   shared_fine_points[1:, 1:, :-1]
                fine_cp[fine_k_base : fine_k_base + k_ratio, fine_j_base : fine_j_base + j_ratio, fine_i_base : fine_i_base + i_ratio, 1, 1, 1] =  \
                   shared_fine_points[1:, 1:, 1:]

    return rqi.grid_from_cp(model,
                            fine_cp,
                            source_grid.crs_uuid,
                            ijk_handedness = 'right' if source_grid.grid_is_right_handed else 'left')


def _refined_unfaulted_grid(model, source_grid, fine_coarse):

    source_points = source_grid.points_ref()
    assert source_points is not None, 'geometry not available for refinement of unfaulted grid'

    # create a new, empty grid object
    grid = grr.Grid(model)

    # inherit attributes from source grid
    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = fine_coarse.fine_extent_kji
    grid.nk, grid.nj, grid.ni = grid.extent_kji[0], grid.extent_kji[1], grid.extent_kji[2]
    grid.k_direction_is_down = source_grid.k_direction_is_down
    grid.grid_is_right_handed = source_grid.grid_is_right_handed
    grid.pillar_shape = source_grid.pillar_shape
    grid.has_split_coordinate_lines = source_grid.has_split_coordinate_lines
    grid.split_pillars_count = source_grid.split_pillars_count
    grid.k_gaps = source_grid.k_gaps
    if grid.k_gaps:
        grid.k_gap_after_array = np.zeros((grid.nk - 1,), dtype = bool)
        grid.k_raw_index_array = np.zeros((grid.nk,), dtype = int)
        # k gap arrays populated below
    # inherit the coordinate reference system used by the grid geometry
    grid.crs_uuid = source_grid.crs_uuid
    if source_grid.model is not model:
        model.duplicate_node(source_grid.model.root_for_uuid(source_grid, grid.crs_uuid), add_as_part = True)
    grid.crs = rqc.Crs(model, uuid = grid.crs_uuid)

    refined_points = np.empty((grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1, 3))

    # log.debug(f'source grid: {source_grid.extent_kji}; k gaps: {source_grid.k_gaps}')
    # log.debug(f'refined grid: {grid.extent_kji}; k gaps: {grid.k_gaps}')
    fk0 = 0
    gaps_so_far = 0
    for ck0 in range(fine_coarse.coarse_extent_kji[0] + 1):
        end_k = (ck0 == fine_coarse.coarse_extent_kji[0])
        if end_k:
            k_ratio = 1
            k_interpolation = [0.0]
        else:
            k_ratio = fine_coarse.ratio(0, ck0)
            k_interpolation = fine_coarse.interpolation(0, ck0)
        one_if_gap = 1 if source_grid.k_gaps and ck0 < fine_coarse.coarse_extent_kji[
            0] - 1 and source_grid.k_gap_after_array[ck0] else 0
        for flk0 in range(k_ratio + one_if_gap):
            # log.debug(f'ck0: {ck0}; fk0: {fk0}; flk0: {flk0}; k_ratio: {k_ratio}; one_if_gap: {one_if_gap}; gaps so far: {gaps_so_far}')
            if flk0 < k_ratio:
                k_fraction = k_interpolation[flk0]
            else:
                k_fraction = 1.0
            if grid.k_gaps:
                if end_k:
                    k_plane = source_points[source_grid.k_raw_index_array[ck0 - 1] + 1, :, :, :]
                else:
                    k_plane = (k_fraction * source_points[source_grid.k_raw_index_array[ck0] + 1, :, :, :] +
                               (1.0 - k_fraction) * source_points[source_grid.k_raw_index_array[ck0], :, :, :])
                if flk0 == k_ratio:
                    grid.k_gap_after_array[fk0 - 1] = True
                elif fk0 < grid.nk:
                    grid.k_raw_index_array[fk0] = fk0 + gaps_so_far
            else:
                if end_k:
                    k_plane = source_points[ck0, :, :, :]
                else:
                    k_plane = k_fraction * source_points[ck0 + 1, :, :, :] + (1.0 -
                                                                              k_fraction) * source_points[ck0, :, :, :]
            fj0 = 0
            for cj0 in range(fine_coarse.coarse_extent_kji[1] + 1):
                end_j = (cj0 == fine_coarse.coarse_extent_kji[1])
                if end_j:
                    j_ratio = 1
                    j_interpolation = [0.0]
                else:
                    j_ratio = fine_coarse.ratio(1, cj0)
                    j_interpolation = fine_coarse.interpolation(1, cj0)
                for flj0 in range(j_ratio):
                    j_fraction = j_interpolation[flj0]
                    # note: shape of j_line will be different if there are split pillars in play
                    if end_j:
                        j_line = k_plane[cj0, :, :]
                    else:
                        j_line = j_fraction * k_plane[cj0 + 1, :, :] + (1.0 - j_fraction) * k_plane[cj0, :, :]

                    fi0 = 0
                    for ci0 in range(fine_coarse.coarse_extent_kji[2] + 1):
                        end_i = (ci0 == fine_coarse.coarse_extent_kji[2])
                        if end_i:
                            i_ratio = 1
                            i_interpolation = [0.0]
                        else:
                            i_ratio = fine_coarse.ratio(2, ci0)
                            i_interpolation = fine_coarse.interpolation(2, ci0)
                        for fli0 in range(i_ratio):
                            i_fraction = i_interpolation[fli0]
                            if end_i:
                                p = j_line[ci0, :]
                            else:
                                p = i_fraction * j_line[ci0 + 1, :] + (1.0 - i_fraction) * j_line[ci0, :]

                            refined_points[fk0 + gaps_so_far, fj0, fi0] = p

                            fi0 += 1

                    assert fi0 == fine_coarse.fine_extent_kji[2] + 1

                    fj0 += 1

            assert fj0 == fine_coarse.fine_extent_kji[1] + 1

            if flk0 == k_ratio:
                gaps_so_far += 1
            else:
                fk0 += 1

    assert fk0 == fine_coarse.fine_extent_kji[0] + 1
    assert grid.nk + gaps_so_far == grid.nk_plus_k_gaps

    grid.points_cached = refined_points

    grid.geometry_defined_for_all_pillars_cached = True
    grid.geometry_defined_for_all_cells_cached = True
    grid.array_cell_geometry_is_defined = np.full(tuple(grid.extent_kji), True, dtype = bool)

    return grid


def _inherit_properties(source_grid, grid, fine_coarse, inherit_realization, inherit_all_realizations):
    source_collection = source_grid.extract_property_collection()
    collection = None
    if source_collection is not None:
        #  do not inherit the inactive property array by this mechanism
        collection = rqp.GridPropertyCollection()
        collection.set_grid(grid)
        collection.extend_imported_list_copying_properties_from_other_grid_collection(
            source_collection,
            refinement = fine_coarse,
            realization = inherit_realization,
            copy_all_realizations = inherit_all_realizations)
    return collection


def _set_parent_window(set_parent_window, source_grid, grid, fine_coarse):
    pw_grid_uuid = source_grid.uuid
    if isinstance(set_parent_window, str):
        if set_parent_window == 'grandparent':
            assert fine_coarse.within_coarse_box is None or (np.all(fine_coarse.within_coarse_box[0] == 0) and
                                                             np.all(fine_coarse.within_coarse_box[1]) == source_grid.extent_kji - 1),  \
               'attempt to set grandparent window for grid when parent window is present'
            source_fine_coarse = source_grid.parent_window
            if source_fine_coarse is not None and (source_fine_coarse.within_fine_box is not None or
                                                   source_fine_coarse.within_coarse_box is not None):
                assert source_fine_coarse.fine_extent_kji == source_fine_coarse.coarse_extent_kji, 'parentage involves refinement or coarsening'
                if source_fine_coarse.within_coarse_box is not None:
                    fine_coarse.within_coarse_box = source_fine_coarse.within_coarse_box
                else:
                    fine_coarse.within_coarse_box = source_fine_coarse.within_fine_box
                pw_grid_uuid = bu.uuid_from_string(
                    rqet.find_nested_tags_text(source_grid.root, ['ParentWindow', 'ParentGrid', 'UUID']))
        else:
            assert set_parent_window == 'parent', 'set_parent_window value not recognized: ' + set_parent_window
    grid.set_parent(pw_grid_uuid, True, fine_coarse)


def _establish_models_and_source_grid(epc_file, new_epc_file, source_grid, source_grid_uuid):
    assert epc_file or source_grid is not None, 'neither epc file name nor source grid supplied'
    if not epc_file:
        epc_file = source_grid.model.epc_file
        assert epc_file, 'unable to ascertain epc filename from grid object'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model = None
    if new_epc_file:
        log.debug('creating fresh model for refined grid')
        model = rq.Model(epc_file = new_epc_file, new_epc = True, create_basics = True, create_hdf5_ext = True)
    if epc_file:
        model_in = rq.Model(epc_file)
        if source_grid is None:
            if source_grid_uuid is None:
                log.debug('using default source grid from existing epc')
                source_grid = model_in.grid()
            else:
                log.debug('selecting source grid from existing epc based on uuid')
                source_grid = grr.Grid(model_in, uuid = source_grid_uuid)
        else:
            if source_grid_uuid is not None:
                assert bu.matching_uuids(source_grid_uuid, source_grid.uuid)
            grid_uuid = source_grid.uuid
            log.debug('reloading source grid from existing epc file')
            source_grid = grr.Grid(model_in, uuid = grid_uuid)
        if model is None:
            model = model_in
    else:
        model_in = source_grid.model
    assert model_in is not None
    assert model is not None
    assert source_grid is not None
    assert source_grid.grid_representation in ['IjkGrid', 'IjkBlockGrid']
    return epc_file, model, model_in, source_grid



"""RESQML coordinate reference systems."""

import logging

log = logging.getLogger(__name__)

import math as maths
import uuid
import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.weights_and_measures as wam
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns

PointType = Union[Tuple[float, float, float], List[float], np.ndarray]


class Crs(BaseResqpy):
    """Coordinate reference system object."""

    @property
    def resqml_type(self):
        """Returns the RESQML class for this crs."""

        return 'LocalTime3dCrs' if hasattr(self, 'time_units') and self.time_units else 'LocalDepth3dCrs'

    valid_axis_orders = ("easting northing", "northing easting", "westing southing", "southing westing",
                         "northing westing", "westing northing")

    def __init__(self,
                 parent_model: 'rq.Model',
                 uuid: Optional[uuid.UUID] = None,
                 x_offset: float = 0.0,
                 y_offset: float = 0.0,
                 z_offset: float = 0.0,
                 rotation: float = 0.0,
                 rotation_units: str = 'dega',
                 xy_units: str = 'm',
                 z_units: str = 'm',
                 z_inc_down: bool = True,
                 axis_order: str = 'easting northing',
                 time_units: Optional[str] = None,
                 epsg_code: Optional[str] = None,
                 title: Optional[str] = None,
                 originator: Optional[str] = None,
                 extra_metadata: Optional[Dict[str, str]] = None):
        """Create a new coordinate reference system object.

        arguments:
            parent_model (model.Model): the model to which the new Crs object will belong
            crs_uuid (uuid.UUID): the uuid of an existing RESQML crs object in model, from which to instantiate the
                resqpy Crs object; if present, all the remaining arguments are ignored
            x_offset, y_offset, z_offset (floats, default zero): the local origin within an implicit parent crs
            rotation (float, default zero): a projected view rotation (in the xy plane), in radians, relative to an
                implicit parent crs; positive value indicates local y axis is clockwise from global y axis.
            rotation_units (str, default 'dega'): the units applicable to rotation, usually 'dega' or 'rad'
            xy_units (str, default 'm'): the units applicable to x & y values; must be one of the RESQML length uom strings
            z_units (str, default 'm'): the units applicable to z depth values; must be one of the RESQML length uom strings
            z_inc_down (boolean, default True): if True, increasing z values indicate greater depth (ie. in direction of
                gravity); if False, increasing z values indicate greater elevation
            axis_order (str, default 'easting northing'): the compass directions of the positive x and y axes respectively
            time_units (str, optional): if present, the time units of the z values (for seismic datasets), in which case the
                z_units argument is irrelevant
            epsg_code (str, optional): if present, the EPSG code of the implicit parent crs
            title (str, optional): the citation title to use for a new crs;
                ignored if uuid is not None
            originator (str, optional): the name of the person creating the crs, defaults to login id;
                ignored if uuid is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the crs;
                ignored if uuid is not None

        returns:
            a new resqpy Crs object

        notes:
            although resqpy does not have full support for identifying a parent crs, the modifiers such as a local origin
            may be used on the assumption that all crs objects refer to the same implicit parent crs;
            there are two equivalent RESQML classes and which one is generated depends on whether the time_units argument
            is used, when instantiating from values;
            it is strongly encourage to call the create_xml() method immediately after instantiation (unless the resqpy
            crs is a temporary object) as the uuid may be modified at that point to re-use any existing equivalent RESQML
            crs object in the model;
            a positive rotation implies that local coordinates are clockwise from global coordinates

        :meta common:
        """

        self.xy_units = xy_units
        self.z_units = z_units
        self.time_units = time_units  # if None, z values are depth; if not None, z values are time (from seismic)
        self.z_inc_down = z_inc_down
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.z_offset = z_offset
        self.rotation = rotation
        self.rotation_units = rotation_units
        self.axis_order = axis_order
        self.epsg_code = epsg_code
        # following are derived attributes, set below
        self.rotated = None
        self.rotation_matrix = None
        self.reverse_rotation_matrix = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        assert self.xy_units in wam.valid_uoms(quantity = 'length'), f'invalid CRS xy units: {self.xy_units}'
        assert self.z_units in wam.valid_uoms(quantity = 'length'), f'invalid CRS z units: {self.z_units}'
        assert self.rotation_units in wam.valid_uoms(quantity='plane angle'), \
            f'invalid CRS rotation units: {self.rotation_units}'
        if self.time_units:
            assert self.time_units in wam.valid_uoms(quantity = 'time'), f'invalid CRS time units: {self.time_units}'
        assert self.axis_order in self.valid_axis_orders, 'invalid CRS axis order: ' + str(axis_order)

        self.set_rotation_matrices()

    def _load_from_xml(self):
        root_node = self.root
        assert root_node is not None
        flavour = rqet.node_type(root_node)
        assert flavour in ['obj_LocalDepth3dCrs', 'obj_LocalTime3dCrs'], f'bad crs node type: {flavour}'
        if flavour == 'obj_LocalTime3dCrs':
            self.time_units = rqet.find_tag_text(root_node, 'TimeUom')
            assert self.time_units
        else:
            self.time_units = None
        self.xy_units = rqet.find_tag_text(root_node, 'ProjectedUom')
        self.axis_order = rqet.find_tag_text(root_node, 'ProjectedAxisOrder')
        self.z_units = rqet.find_tag_text(root_node, 'VerticalUom')
        self.z_inc_down = rqet.find_tag_bool(root_node, 'ZIncreasingDownward')
        self.x_offset = rqet.find_tag_float(root_node, 'XOffset')
        self.y_offset = rqet.find_tag_float(root_node, 'YOffset')
        self.z_offset = rqet.find_tag_float(root_node, 'ZOffset')
        self.rotation = rqet.find_tag_float(root_node, 'ArealRotation')
        rotation_node = rqet.find_tag(root_node, 'ArealRotation')
        self.rotation_units = rotation_node.attrib.get('uom')
        parent_xy_crs = rqet.find_tag(root_node, 'ProjectedCrs')
        if parent_xy_crs is not None and rqet.node_type(parent_xy_crs) == 'ProjectedCrsEpsgCode':
            self.epsg_code = rqet.find_tag_text(parent_xy_crs, 'EpsgCode')  # should be an integer?
        else:
            self.epsg_code = None

    def set_rotation_matrices(self):
        """Sets the rotation matrices, and the rotated and null_transform flags, call after changing rotation."""

        rotation_deg = wam.convert(self.rotation, self.rotation_units, 'dega', quantity = 'plane angle')
        # log.debug(f'setting rotation matrices for dega: {rotation_deg}')
        self.rotated = (not maths.isclose(rotation_deg, 0.0, abs_tol = 1e-8) and
                        not maths.isclose(rotation_deg, 360.0, abs_tol = 1e-8))
        if self.rotated:
            self.rotation_matrix = vec.rotation_matrix_3d_axial(2, rotation_deg)
            self.reverse_rotation_matrix = vec.rotation_matrix_3d_axial(2, -rotation_deg)
            if self.is_right_handed_xy():
                self.rotation_matrix, self.reverse_rotation_matrix = self.reverse_rotation_matrix, self.rotation_matrix
        else:
            self.rotation_matrix = None
            self.reverse_rotation_matrix = None
        self.null_transform = (maths.isclose(self.x_offset, 0.0, abs_tol = 1e-8) and
                               maths.isclose(self.y_offset, 0.0, abs_tol = 1e-8) and
                               maths.isclose(self.z_offset, 0.0, abs_tol = 1e-8) and not self.rotated)

    def is_right_handed_xy(self) -> bool:
        """Returns True if the xy axes are right handed when viewed from above; False if left handed."""

        return bool(self.axis_order in ["northing easting", "southing westing", "westing northing"])

    def is_right_handed_xyz(self) -> bool:
        """Returns True if the xyz axes are right handed; False if left handed."""

        return self.is_right_handed_xy() is bool(self.z_inc_down)

    def global_to_local(self, xyz: PointType, global_z_inc_down: bool = True) -> Tuple[float, float, float]:
        """Convert a single xyz point from the parent coordinate reference system to this one."""

        x, y, z = xyz
        if self.x_offset != 0.0:
            x -= self.x_offset
        if self.y_offset != 0.0:
            y -= self.y_offset
        if global_z_inc_down != self.z_inc_down:
            z = -z
        if self.z_offset != 0.0:
            z -= self.z_offset
        if self.rotated:
            (x, y, z) = vec.rotate_vector(self.rotation_matrix, np.array((x, y, z)))
        return (x, y, z)

    def global_to_local_array(self, xyz: np.ndarray, global_z_inc_down: bool = True):
        """Convert in situ a numpy array of xyz points from the parent coordinate reference system to this one."""

        if self.x_offset != 0.0:
            xyz[..., 0] -= self.x_offset
        if self.y_offset != 0.0:
            xyz[..., 1] -= self.y_offset
        if global_z_inc_down != self.z_inc_down:
            z = np.negative(xyz[..., 2])
            xyz[..., 2] = z
        if self.z_offset != 0.0:
            xyz[..., 2] -= self.z_offset
        if self.rotated:
            a = vec.rotate_array(self.rotation_matrix, xyz)
            xyz[:] = a

    def local_to_global(self, xyz: PointType, global_z_inc_down: bool = True) -> Tuple[float, float, float]:
        """Convert a single xyz point from this coordinate reference system to the parent one."""

        if self.rotated:
            (x, y, z) = vec.rotate_vector(self.reverse_rotation_matrix, np.array(xyz))
        else:
            (x, y, z) = xyz
        if self.x_offset != 0.0:
            x += self.x_offset
        if self.y_offset != 0.0:
            y += self.y_offset
        if self.z_offset != 0.0:
            z += self.z_offset
        if global_z_inc_down != self.z_inc_down:
            z = -z
        return (x, y, z)

    def local_to_global_array(self, xyz: np.ndarray, global_z_inc_down: bool = True):
        """Convert in situ a numpy array of xyz points from this coordinate reference system to the parent one."""

        if self.rotated:
            a = vec.rotate_array(self.reverse_rotation_matrix, xyz)
            xyz[:] = a
        if self.x_offset != 0.0:
            xyz[..., 0] += self.x_offset
        if self.y_offset != 0.0:
            xyz[..., 1] += self.y_offset
        if self.z_offset != 0.0:
            xyz[..., 2] += self.z_offset
        if global_z_inc_down != self.z_inc_down:
            z = np.negative(xyz[..., 2])
            xyz[..., 2] = z

    def has_same_epsg_code(self, other_crs: 'Crs') -> bool:
        """Returns True if either of the crs'es has a null EPSG code, or if they are the same."""
        return not self.epsg_code or not other_crs.epsg_code or self.epsg_code == other_crs.epsg_code

    def is_equivalent(self, other_crs: 'Crs') -> bool:
        """Returns True if this crs is effectively the same as the other crs."""

        # log.debug('testing crs equivalence')
        if other_crs is None:
            return False
        if self is other_crs:
            return True
        if bu.matching_uuids(self.uuid, other_crs.uuid):
            return True
        if self.resqml_type != other_crs.resqml_type:
            return False
        if self.xy_units != other_crs.xy_units or self.z_units != other_crs.z_units:
            return False
        if self.z_inc_down != other_crs.z_inc_down:
            return False
        if (self.time_units is not None or
                other_crs.time_units is not None) and self.time_units != other_crs.time_units:
            return False
        if self.axis_order != other_crs.axis_order:
            return False
        if not self.has_same_epsg_code(other_crs):
            return False
        if self.null_transform and other_crs.null_transform:
            return True
        if (maths.isclose(self.x_offset, other_crs.x_offset, abs_tol = 1e-4) and
                maths.isclose(self.y_offset, other_crs.y_offset, abs_tol = 1e-4) and
                maths.isclose(self.z_offset, other_crs.z_offset, abs_tol = 1e-4) and
                maths.isclose(self.rotation, other_crs.rotation, abs_tol = 1e-4)):
            return True
        # todo: handle and check rotation units; modularly equivalent rotations
        return False

    def convert_to(self, other_crs: 'Crs', xyz: PointType) -> Tuple[float, float, float]:
        """Converts a single xyz point from this coordinate reference system to the other.

        :meta common:
        """

        if self is other_crs:
            return _as_xyz_tuple(xyz)
        assert self.resqml_type == other_crs.resqml_type
        # assert self.has_same_epsg_code(other_crs)
        if not self.has_same_epsg_code(other_crs):
            log.warning("converting between crs'es with different epsg codes")
        xyz = self.local_to_global(xyz)
        # yapf: disable
        if self.resqml_type == 'LocalDepth3dCrs':
            xyz = (wam.convert_lengths(xyz[0], self.xy_units, other_crs.xy_units),
                   wam.convert_lengths(xyz[1], self.xy_units, other_crs.xy_units),
                   wam.convert_lengths(xyz[2], self.z_units, other_crs.z_units))
        else:
            xyz = (wam.convert_lengths(xyz[0], self.xy_units, other_crs.xy_units),
                   wam.convert_lengths(xyz[1], self.xy_units, other_crs.xy_units),
                   wam.convert_times(xyz[2], self.time_units, other_crs.time_units))
        # yapf: enable
        xyz = other_crs.global_to_local(xyz)
        return _as_xyz_tuple(xyz)

    def convert_array_to(self, other_crs: 'Crs', xyz: np.ndarray):
        """Converts in situ a numpy array of xyz points from this coordinate reference system to the other.

        :meta common:
        """

        if self.is_equivalent(other_crs):
            return xyz
        assert self.resqml_type == other_crs.resqml_type
        # assert self.has_same_epsg_code(other_crs)
        if not self.has_same_epsg_code(other_crs):
            log.warning("converting between crs'es with different epsg codes")
        self.local_to_global_array(xyz)
        if self.resqml_type == 'LocalDepth3dCrs':
            if self.xy_units == self.z_units and other_crs.xy_units == other_crs.z_units:
                wam.convert_lengths(xyz, self.xy_units, other_crs.xy_units)
            else:
                wam.convert_lengths(xyz[..., :2], self.xy_units, other_crs.xy_units)
                wam.convert_lengths(xyz[..., 2], self.z_units, other_crs.z_units)
        else:
            wam.convert_lengths(xyz[..., :2], self.xy_units, other_crs.xy_units)
            wam.convert_times(xyz[..., 2], self.time_units, other_crs.time_units)
        other_crs.global_to_local_array(xyz)
        return xyz

    def convert_from(self, other_crs: 'Crs', xyz: PointType) -> Tuple[float, float, float]:
        """Converts a single xyz point from the other coordinate reference system to this one.

        :meta common:
        """

        if self is other_crs:
            return _as_xyz_tuple(xyz)
        assert self.resqml_type == other_crs.resqml_type
        # assert self.has_same_epsg_code(other_crs)
        if not self.has_same_epsg_code(other_crs):
            log.warning("converting between crs'es with different epsg codes")
        xyz = other_crs.local_to_global(xyz)
        # yapf: disable
        if self.resqml_type == 'LocalDepth3dCrs':
            xyz = (wam.convert_lengths(xyz[0], other_crs.xy_units, self.xy_units),
                   wam.convert_lengths(xyz[1], other_crs.xy_units, self.xy_units),
                   wam.convert_lengths(xyz[2], other_crs.z_units, self.z_units))
        else:
            xyz = (wam.convert_lengths(xyz[0], other_crs.xy_units, self.xy_units),
                   wam.convert_lengths(xyz[1], other_crs.xy_units, self.xy_units),
                   wam.convert_times(xyz[2], other_crs.time_units, self.time_units))
        # yapf: enable
        xyz = self.global_to_local(xyz)
        return _as_xyz_tuple(xyz)

    def convert_array_from(self, other_crs: 'Crs', xyz: np.ndarray):
        """Converts in situ a numpy array of xyz points from the other coordinate reference system to this one.

        :meta common:
        """

        if self.is_equivalent(other_crs):
            return xyz
        assert self.resqml_type == other_crs.resqml_type
        # assert self.has_same_epsg_code(other_crs)
        if not self.has_same_epsg_code(other_crs):
            log.warning("converting between crs'es with different epsg codes")
        other_crs.local_to_global_array(xyz)
        if self.resqml_type == 'LocalDepth3dCrs':
            if self.xy_units == self.z_units and other_crs.xy_units == other_crs.z_units:
                wam.convert_lengths(xyz, other_crs.xy_units, self.xy_units)
            else:
                wam.convert_lengths(xyz[..., :2], other_crs.xy_units, self.xy_units)
                wam.convert_lengths(xyz[..., 2], other_crs.z_units, self.z_units)
        else:
            wam.convert_lengths(xyz[..., :2], other_crs.xy_units, self.xy_units)
            wam.convert_times(xyz[..., 2], other_crs.time_units, self.time_units)
        self.global_to_local_array(xyz)
        return xyz

    def create_xml(self,
                   title: Optional[str] = None,
                   originator: Optional[str] = None,
                   extra_metadata: Optional[Dict[str, str]] = None,
                   add_as_part: bool = True,
                   reuse: bool = True):
        """Creates a Coordinate Reference System xml node and optionally adds as a part in the parent model.

        arguments:
            title (string, optional): used as the Title text in the citation node
            originator (string, optional): the name of the human being who created the crs object;
                default is to use the login name
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the crs
            add_as_part (boolean, default True): if True the newly created crs node is added to the model
                as a part
            reuse (boolean, default True): if True and an equivalent crs already exists in the model then
                the uuid for this Crs is modified to match that of the existing object and the existing
                xml node is returned without anything new being added

        returns:
            newly created (or reused) coordinate reference system xml node

        notes:
            if the reuse argument is True, it is strongly recommended to call this method immediately after
            a new Crs has been instantiated from explicit values, as the uuid may be modified here;
            if reuse is True, the title is not regarded as important and a match with an existing object
            may occur even if the titles differ

        :meta common:
        """

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object

        crs = super().create_xml(add_as_part = False,
                                 title = title,
                                 originator = originator,
                                 extra_metadata = extra_metadata)

        xoffset = rqet.SubElement(crs, ns['resqml2'] + 'XOffset')
        xoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        xoffset.text = '{0:5.3f}'.format(self.x_offset)

        yoffset = rqet.SubElement(crs, ns['resqml2'] + 'YOffset')
        yoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        yoffset.text = '{0:5.3f}'.format(self.y_offset)

        zoffset = rqet.SubElement(crs, ns['resqml2'] + 'ZOffset')
        zoffset.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        zoffset.text = '{0:6.4f}'.format(self.z_offset)

        rotation = rqet.SubElement(crs, ns['resqml2'] + 'ArealRotation')
        rotation.set('uom', self.rotation_units)
        rotation.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
        rotation.text = '{0:8.6f}'.format(self.rotation)

        if self.axis_order is None:
            axes = 'easting northing'
        else:
            axes = self.axis_order.lower()
        axis_order = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedAxisOrder')
        axis_order.set(ns['xsi'] + 'type', ns['eml'] + 'AxisOrder2d')
        axis_order.text = axes

        xy_uom = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedUom')
        xy_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
        xy_uom.text = wam.rq_length_unit(self.xy_units)

        z_uom = rqet.SubElement(crs, ns['resqml2'] + 'VerticalUom')
        z_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
        z_uom.text = wam.rq_length_unit(self.z_units)

        if self.time_units is not None:
            t_uom = rqet.SubElement(crs, ns['resqml2'] + 'TimeUom')
            t_uom.set(ns['xsi'] + 'type', ns['eml'] + 'TimeUom')  # todo: check this
            t_uom.text = self.time_units

        z_sense = rqet.SubElement(crs, ns['resqml2'] + 'ZIncreasingDownward')
        z_sense.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        z_sense.text = str(self.z_inc_down).lower()

        xy_crs = rqet.SubElement(crs, ns['resqml2'] + 'ProjectedCrs')
        xy_crs.text = rqet.null_xml_text
        if self.epsg_code is None:
            xy_crs.set(ns['xsi'] + 'type', ns['eml'] + 'ProjectedUnknownCrs')
            self.model.create_unknown(root = xy_crs)
        else:
            xy_crs.set(ns['xsi'] + 'type', ns['eml'] + 'ProjectedCrsEpsgCode')
            epsg_node = rqet.SubElement(xy_crs, ns['resqml2'] + 'EpsgCode')
            epsg_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            epsg_node.text = str(self.epsg_code)

        z_crs = rqet.SubElement(crs, ns['resqml2'] + 'VerticalCrs')
        if self.epsg_code is None:
            z_crs.set(ns['xsi'] + 'type', ns['eml'] + 'VerticalUnknownCrs')
            z_crs.text = rqet.null_xml_text
            self.model.create_unknown(root = z_crs)
        else:  # not sure if this is appropriate for the vertical crs
            z_crs.set(ns['xsi'] + 'type', ns['eml'] + 'VerticalCrsEpsgCode')
            epsg_node = rqet.SubElement(z_crs, ns['resqml2'] + 'EpsgCode')
            epsg_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            epsg_node.text = str(self.epsg_code)

        if add_as_part:
            self.model.add_part('obj_' + self.resqml_type, bu.uuid_from_string(crs.attrib['uuid']), crs)
        if self.model.crs_uuid is None:
            self.model.crs_uuid = self.uuid  # mark's as 'main' (ie. first) crs for model

        return crs


def _as_xyz_tuple(xyz):
    """Coerce into 3-tuple of floats."""

    return tuple((float(xyz[0]), float(xyz[1]), float(xyz[2])))



"""Factorization and functions supporting grid extent determination from corner points."""

version = '24th December 2019'


def factorize(n):
    """Returns list of prime factors of positive integer n."""
    i = 2
    factors = []
    while True:
        q, r = divmod(n, i)
        if r:
            i += 1
            if i > n:
                return factors
        else:
            factors.append(i)
            if q < i:
                return factors
            n = q


def combinatorial(numbers):
    """Returns a list of all possible product combinations of numbers from list numbers, with some duplicates."""
    if len(numbers) == 0:
        return []
    head = numbers[0]
    c = [head]
    tail = combinatorial(numbers[1:])
    for o in tail:
        if o != head:
            c.append(o)
        c.append(head * o)
    return sorted(c)


def all_factors_from_primes(primes):
    """Returns a sorted list of unique factors from prime factorization."""
    all = list(set(combinatorial(primes)))
    all.append(1)
    return sorted(all)


def all_factors(n):
    """Returns a sorted list of unique factors of n."""
    primes = factorize(n)
    return all_factors_from_primes(primes)


def remove_subset(primary, subset):
    """Remove all elements of subset from primary list."""

    for e in subset:
        primary.remove(e)



"""_wellbore_marker.py: resqpy well module providing wellbore marker class"""

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.xml_et as rqet
from resqpy.olio.xml_namespaces import curly_namespace as ns
import resqpy.olio.uuid as bu


class WellboreMarker():
    """Class to handle RESQML WellboreMarker objects.

    note:
       wellbore markers are not high level RESQML objects
    """

    resqml_type = 'WellboreMarker'

    boundary_feature_dict = {
        'GeologicBoundaryKind': ['fault', 'geobody', 'horizon'],
        'FluidMarker': ['gas down to', 'gas up to', 'oil down to', 'oil up to', 'water down to', 'water up to'],
        'FluidContact': ['free water contact', 'gas oil contact', 'gas water contact', 'seal', 'water oil contact']
    }

    def __init__(self,
                 parent_model,
                 parent_frame,
                 marker_index,
                 marker_node = None,
                 marker_type = None,
                 interpretation_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Creates a new wellbore marker object and optionally loads it from xml.

        arguments:
           parent_model (model.Model object): the model which the new wellbore marker belongs to
           parent_frame (wellbore_marker_frame.WellboreMarkerFramer object): the wellbore marker frame to which the wellbore marker belongs
           marker_index (int): index of the wellbore marker in the parent WellboreMarkerFrame object
           marker_node (xml node, optional): if given, loads from xml. Else, creates new
           marker_type (str, optional): the type of geologic, fluid or contact feature
             e.g. "fault", "geobody", "horizon ", "gas/oil/water down to", "gas/oil/water up to",
              "free water contact", "gas oil contact", "gas water contact", "water oil contact", "seal"
           interpretation_uuid (uuid.UUID or string, optional): uuid of the boundary feature Interpretation
              organizational object that the marker refers to.
              note: it is highly recommended that a related boundary feature interpretation is provided
           title (str, optional): the citation title to use for a new wellbore marker
              ignored if uuid is not None;
           originator (str, optional): the name of the person creating the wellbore marker, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the wellbore marker;
              ignored if uuid is not None

        returns:
           the newly created wellbore marker object
        """
        # verify that marker type is valid
        if marker_type is not None:
            assert marker_type in ([
                "fault", "geobody", "horizon", "gas down to", "oil down to", "water down to", "gas up to", "oil up to",
                "water up to", "free water contact", "gas oil contact", "gas water contact", "water oil contact", "seal"
            ]), "invalid marker type specified"

        self.model = parent_model
        self.wellbore_frame = parent_frame
        self.marker_index = marker_index
        self.uuid = None
        self.marker_type = marker_type
        self.interpretation_uuid = interpretation_uuid
        self.title = title
        self.originator = originator
        self.extra_metadata = extra_metadata
        if marker_node is not None:
            self._load_from_xml(marker_node = marker_node)
        if self.uuid is None:
            self.uuid = bu.new_uuid()
        assert self.uuid is not None

    def create_xml(self, parent_node, title = 'wellbore marker'):
        """Creates the xml tree for this wellbore marker.

        arguments:
           parent_node (xml node): the root node of the WellboreMarkerFrame object to which the newly created node will be appended
           title (string, optional): the citation title of the newly created node
             note: if not None, self.title will be used instead of "wellbore marker"

        returns:
           the newly created xml node
        """

        assert self.uuid is not None
        wbm_node = self.model.new_obj_node('WellboreMarker', is_top_lvl_obj = False)
        wbm_node.set('uuid', str(self.uuid))

        # Citation block
        if self.title is not None:
            title = self.title
        citation = self.model.create_citation(root = wbm_node, title = title, originator = self.originator)

        # Add sub-elements to root node
        for k, v in WellboreMarker.boundary_feature_dict.items():
            if self.marker_type in v:
                boundary_kind = k
                break

        wbm_gb_node = rqet.SubElement(wbm_node, ns['resqml2'] + boundary_kind)
        # wbm_gb_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
        wbm_gb_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicBoundaryKind')
        wbm_gb_node.text = str(self.marker_type)

        if self.interpretation_uuid is not None:
            interp_content_type = 'obj_' + self.marker_type.capitalize() + 'Interpretation'
            interp_root = self.model.root_for_uuid(uuid = self.interpretation_uuid)
            self.model.create_ref_node(flavour = 'Interpretation',
                                       title = rqet.find_tag(rqet.find_tag(interp_root, 'Citation'), 'Title').text,
                                       uuid = self.interpretation_uuid,
                                       content_type = interp_content_type,
                                       root = wbm_node)
        # Extra metadata
        if hasattr(self, 'extra_metadata') and self.extra_metadata:
            rqet.create_metadata_xml(node = wbm_node, extra_metadata = self.extra_metadata)

        if parent_node is not None:
            parent_node.append(wbm_node)

        return wbm_node

    def _load_from_xml(self, marker_node):
        """Load attributes from xml.

        This is invoked as part of the init method when an existing uuid is given.

        Returns:
           [bool]: True if successful
        """

        assert marker_node is not None

        # Load XML data
        uuid_str = marker_node.attrib.get('uuid')
        if uuid_str:
            self.uuid = bu.uuid_from_string(uuid_str)
        citation_tag = rqet.find_nested_tags(root = marker_node, tag_list = ['Citation'])
        assert citation_tag is not None
        self.title = rqet.find_tag_text(root = citation_tag, tag_name = 'Title')
        self.originator = rqet.find_tag_text(root = citation_tag, tag_name = 'Originator')

        self.marker_type = None
        for boundary_feature_type in ['GeologicBoundaryKind', 'FluidMarker', 'FluidContact']:
            found_tag_text = rqet.find_tag_text(root = marker_node, tag_name = boundary_feature_type)
            if found_tag_text is not None:
                self.marker_type = found_tag_text
                break
        assert self.marker_type is not None

        self.interpretation_uuid = bu.uuid_from_string(
            rqet.find_nested_tags_text(root = marker_node, tag_list = ['Interpretation', 'UUID']))
        self.extra_metadata = rqet.load_metadata_from_xml(node = marker_node)

        return True



"""Collections of properties for grids, wellbore frames, grid connection sets etc."""

__all__ = [
    'PropertyCollection', 'Property', 'WellLog', 'WellIntervalProperty', 'WellIntervalPropertyCollection',
    'WellLogCollection', 'StringLookup', 'PropertyKind', 'GridPropertyCollection',
    'property_over_time_series_from_collection', 'property_collection_for_keyword', 'infer_property_kind',
    'write_hdf5_and_create_xml_for_active_property', 'reformat_column_edges_to_resqml_format',
    'reformat_column_edges_from_resqml_format', 'same_property_kind', 'selective_version_of_collection',
    'supported_local_property_kind_list', 'supported_property_kind_list', 'supported_facet_type_list',
    'expected_facet_type_dict', 'create_transmisibility_multiplier_property_kind',
    'property_kind_and_facet_from_keyword', 'guess_uom'
]

from .property_common import property_collection_for_keyword, \
    property_over_time_series_from_collection, \
    write_hdf5_and_create_xml_for_active_property, \
    infer_property_kind, \
    reformat_column_edges_to_resqml_format, \
    reformat_column_edges_from_resqml_format, \
    selective_version_of_collection, \
    same_property_kind, \
    supported_property_kind_list, \
    supported_local_property_kind_list, \
    supported_facet_type_list, \
    expected_facet_type_dict, \
    property_kind_and_facet_from_keyword, \
    guess_uom
from .property_kind import PropertyKind, create_transmisibility_multiplier_property_kind
from .string_lookup import StringLookup
from .property_collection import PropertyCollection
from .grid_property_collection import GridPropertyCollection
from .property import Property
from .well_interval_property import WellIntervalProperty
from .well_interval_property_collection import WellIntervalPropertyCollection
from .well_log import WellLog
from .well_log_collection import WellLogCollection

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""pyramid_grid.py: resqpy PyramidGrid class module."""

version = '24th November 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.grid as grr
import resqpy.olio.volume as vol
from resqpy.unstructured._unstructured_grid import UnstructuredGrid


class PyramidGrid(UnstructuredGrid):
    """Class for unstructured grids where every cell is a quadrilateral pyramid."""

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 cache_geometry = False,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a new resqpy PyramidGrid object (RESQML UnstructuredGrid with cell shape pyramidal)

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created PyramidGrid object
        """

        super().__init__(parent_model = parent_model,
                         uuid = uuid,
                         find_properties = find_properties,
                         geometry_required = True,
                         cache_geometry = cache_geometry,
                         cell_shape = 'pyramidal',
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            assert grr.grid_flavour(self.root) == 'PyramidGrid'
            self.check_pyramidal()

        self.grid_representation = 'PyramidGrid'  #: flavour of grid; not much used

    def check_pyramidal(self):
        """Checks that each cell has 5 faces and each face has 3 or 4 nodes.

        note:
           currently only performs a cursory check, without checking nodes are shared or that there is exactly one
           quadrilateral face
        """

        assert self.cell_shape == 'pyramidal'
        self.cache_all_geometry_arrays()
        assert self.faces_per_cell_cl is not None and self.nodes_per_face_cl is not None
        assert self.faces_per_cell_cl[0] == 5 and np.all(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1] == 5)
        nodes_per_face_count = np.empty(self.face_count)
        nodes_per_face_count[0] = self.nodes_per_face_cl[0]
        nodes_per_face_count[1:] = self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1]
        assert np.all(np.logical_or(nodes_per_face_count == 3, nodes_per_face_count == 4))

    def face_indices_for_cell(self, cell):
        """Returns numpy list of face indices for a single cell.

        arguments:
           cell (int): the index of the cell for which face indices are required

        returns:
           numpy int array of shape (5,) being the face indices of each of the 5 faces for the cell; the first
           index in the array is for the quadrilateral face

        note:
           the face indices are used when accessing the nodes per face data and can also be used to identify
           shared faces
        """

        faces = super().face_indices_for_cell(cell)
        assert len(faces) == 5
        result = -np.ones(5, dtype = int)
        i = 1
        for f in range(5):
            nc = self.node_count_for_face(faces[f])
            if nc == 3:
                assert i < 5, 'too many triangular faces for cell in pyramid grid'
                result[i] = faces[f]
                i += 1
            else:
                assert nc == 4, 'pyramid grid includes a face that is neither triangle nor quadrilateral'
                assert result[0] == -1, 'more than one quadrilateral face for cell in pyramid grid'
                result[0] = faces[f]
        return result

    def face_indices_and_handedness_for_cell(self, cell):
        """Returns numpy list of face indices for a single cell, and numpy boolean list of face right handedness.

        arguments:
           cell (int): the index of the cell for which face indices are required

        returns:
           numpy int array of shape (5,), numpy boolean array of shape (5, ):
           being the face indices of each of the 5 faces for the cell, and the right handedness (clockwise order)
           of the face nodes when viewed from within the cell; the first entry in each list is for the
           quadrilateral face

        note:
           the face indices are used when accessing the nodes per face data and can also be used to identify
           shared faces; the handedness (clockwise or anti-clockwise ordering of nodes) is significant for
           some processing of geometry such as volume calculations
        """

        faces, handednesses = super().face_indices_and_handedness_for_cell(cell)
        assert len(faces) == 5 and len(handednesses) == 5
        f_result = -np.ones(5, dtype = int)
        h_result = np.empty(5, dtype = bool)
        i = 1
        for f in range(5):
            nc = self.node_count_for_face(faces[f])
            if nc == 3:
                assert i < 5, 'too many triangular faces for cell in pyramid grid'
                f_result[i] = faces[f]
                h_result[i] = handednesses[f]
                i += 1
            else:
                assert nc == 4, 'pyramid grid includes a face that is neither triangle nor quadrilateral'
                assert f_result[0] == -1, 'more than one quadrilateral face for cell in pyramid grid'
                f_result[0] = faces[f]
                h_result[0] = handednesses[f]
        return f_result, h_result

    def volume(self, cell):
        """Returns the volume of a single cell.

        arguments:
           cell (int): the index of the cell for which the volume is required

        returns:
           float being the volume of the pyramidal cell; units of measure is implied by crs units
        """

        self._set_crs_handedness()
        self.cache_all_geometry_arrays()
        faces, hands = self.face_indices_and_handedness_for_cell(cell)
        nodes = self.distinct_node_indices_for_cell(cell)
        base_nodes = self.node_indices_for_face(faces[0])
        for node in nodes:
            if node not in base_nodes:
                apex_node = node
                break
        else:
            raise Exception('apex node not found for cell in pyramid grid')
        apex = self.points_cached[apex_node]
        abcd = self.points_cached[base_nodes]

        return vol.pyramid_volume(apex,
                                  abcd[0],
                                  abcd[1],
                                  abcd[2],
                                  abcd[3],
                                  crs_is_right_handed = (self.crs_is_right_handed == hands[0]))

    # todo: add pyramidal specific method for centre_point()



"""Submodule containing functions to create a grid xml file."""

import logging

log = logging.getLogger(__name__)

import resqpy.olio.xml_et as rqet
import resqpy.property as rprop
from resqpy.olio.xml_namespaces import curly_namespace as ns

always_write_pillar_geometry_is_defined_array = False
always_write_cell_geometry_is_defined_array = False


def _create_grid_xml(grid,
                     ijk,
                     ext_uuid = None,
                     add_as_part = True,
                     add_relationships = True,
                     write_active = True,
                     write_geometry = True):
    """Function that returns an xml representation containing grid information"""

    if grid.grid_representation and not write_geometry:
        rqet.create_metadata_xml(node = ijk, extra_metadata = {'grid_flavour': grid.grid_representation})

    ni_node = rqet.SubElement(ijk, ns['resqml2'] + 'Ni')
    ni_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
    ni_node.text = str(grid.extent_kji[2])

    nj_node = rqet.SubElement(ijk, ns['resqml2'] + 'Nj')
    nj_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
    nj_node.text = str(grid.extent_kji[1])

    nk_node = rqet.SubElement(ijk, ns['resqml2'] + 'Nk')
    nk_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
    nk_node.text = str(grid.extent_kji[0])

    if grid.k_gaps:
        kg_node = rqet.SubElement(ijk, ns['resqml2'] + 'KGaps')
        kg_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'KGaps')
        kg_node.text = '\n'

        kgc_node = rqet.SubElement(kg_node, ns['resqml2'] + 'Count')
        kgc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        kgc_node.text = str(grid.k_gaps)

        assert grid.k_gap_after_array.ndim == 1 and grid.k_gap_after_array.size == grid.nk - 1

        kgal_node = rqet.SubElement(kg_node, ns['resqml2'] + 'GapAfterLayer')
        kgal_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanHdf5Array')
        kgal_node.text = '\n'

        kgal_values = rqet.SubElement(kgal_node, ns['resqml2'] + 'Values')
        kgal_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        kgal_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'GapAfterLayer', root = kgal_values)

    if grid.stratigraphic_column_rank_uuid is not None and grid.stratigraphic_units is not None:
        assert grid.model.type_of_uuid(
            grid.stratigraphic_column_rank_uuid) == 'obj_StratigraphicColumnRankInterpretation'

        strata_node = rqet.SubElement(ijk, ns['resqml2'] + 'IntervalStratigraphicUnits')
        strata_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntervalStratigraphicUnits')
        strata_node.text = '\n'

        ui_node = rqet.SubElement(strata_node, ns['resqml2'] + 'UnitIndices')
        ui_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        ui_node.text = '\n'

        ui_null = rqet.SubElement(ui_node, ns['resqml2'] + 'NullValue')
        ui_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        ui_null.text = '-1'

        ui_values = rqet.SubElement(ui_node, ns['resqml2'] + 'Values')
        ui_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        ui_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'unitIndices', root = ui_values)

        grid.model.create_ref_node('StratigraphicOrganization',
                                   grid.model.title(uuid = grid.stratigraphic_column_rank_uuid),
                                   grid.stratigraphic_column_rank_uuid,
                                   content_type = 'StratigraphicColumnRankInterpretation',
                                   root = strata_node)

    if grid.parent_window is not None:

        pw_node = rqet.SubElement(ijk, ns['resqml2'] + 'ParentWindow')
        pw_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IjkParentWindow')
        pw_node.text = '\n'

        assert grid.parent_grid_uuid is not None
        parent_grid_root = grid.model.root(uuid = grid.parent_grid_uuid)
        if parent_grid_root is None:
            pg_title = 'ParentGrid'
        else:
            pg_title = rqet.citation_title_for_node(parent_grid_root)
        grid.model.create_ref_node('ParentGrid',
                                   pg_title,
                                   grid.parent_grid_uuid,
                                   content_type = 'obj_IjkGridRepresentation',
                                   root = pw_node)

        for axis in range(3):

            regrid_node = rqet.SubElement(pw_node, 'KJI'[axis] + 'Regrid')
            regrid_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Regrid')
            regrid_node.text = '\n'

            if grid.is_refinement:
                if grid.parent_window.within_coarse_box is None:
                    iiopg = 0  # InitialIndexOnParentGrid
                else:
                    iiopg = grid.parent_window.within_coarse_box[0, axis]
            else:
                if grid.parent_window.within_fine_box is None:
                    iiopg = 0
                else:
                    iiopg = grid.parent_window.within_fine_box[0, axis]
            iiopg_node = rqet.SubElement(regrid_node, ns['resqml2'] + 'InitialIndexOnParentGrid')
            iiopg_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
            iiopg_node.text = str(iiopg)

            if grid.parent_window.fine_extent_kji[axis] == grid.parent_window.coarse_extent_kji[axis]:
                continue  # one-to-noe mapping

            intervals_node = rqet.SubElement(regrid_node, ns['resqml2'] + 'Intervals')
            intervals_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Intervals')
            intervals_node.text = '\n'

            if grid.parent_window.constant_ratios[axis] is not None:
                interval_count = 1
            else:
                if grid.is_refinement:
                    interval_count = grid.parent_window.coarse_extent_kji[axis]
                else:
                    interval_count = grid.parent_window.fine_extent_kji[axis]
            ic_node = rqet.SubElement(intervals_node, ns['resqml2'] + 'IntervalCount')
            ic_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            ic_node.text = str(interval_count)

            pcpi_node = rqet.SubElement(intervals_node, ns['resqml2'] + 'ParentCountPerInterval')
            pcpi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
            pcpi_node.text = '\n'

            pcpi_values = rqet.SubElement(pcpi_node, ns['resqml2'] + 'Values')
            pcpi_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            pcpi_values.text = '\n'

            grid.model.create_hdf5_dataset_ref(ext_uuid,
                                               grid.uuid,
                                               'KJI'[axis] + 'Regrid/ParentCountPerInterval',
                                               root = pcpi_values)

            ccpi_node = rqet.SubElement(intervals_node, ns['resqml2'] + 'ChildCountPerInterval')
            ccpi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
            ccpi_node.text = '\n'

            ccpi_values = rqet.SubElement(ccpi_node, ns['resqml2'] + 'Values')
            ccpi_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            ccpi_values.text = '\n'

            grid.model.create_hdf5_dataset_ref(ext_uuid,
                                               grid.uuid,
                                               'KJI'[axis] + 'Regrid/ChildCountPerInterval',
                                               root = ccpi_values)

            if grid.is_refinement and not grid.parent_window.equal_proportions[axis]:
                ccw_node = rqet.SubElement(intervals_node, ns['resqml2'] + 'ChildCellWeights')
                ccw_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
                ccw_node.text = rqet.null_xml_text

                ccw_values_node = rqet.SubElement(ccw_node, ns['resqml2'] + 'Values')
                ccw_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
                ccw_values_node.text = rqet.null_xml_text

                grid.model.create_hdf5_dataset_ref(ext_uuid,
                                                   grid.uuid,
                                                   'KJI'[axis] + 'Regrid/ChildCellWeights',
                                                   root = ccw_values_node)

        # todo: handle omit and cell overlap functionality as part of parent window refining or coarsening

    if write_geometry:
        __add_geometry_xml(ext_uuid, grid, ijk)

    if add_as_part:
        __add_as_part(add_relationships, ext_uuid, grid, ijk, write_geometry)

    if (write_active and grid.active_property_uuid is not None and
            grid.model.part(uuid = grid.active_property_uuid) is None):
        # TODO: replace following with call to rprop.write_hdf5_and_create_xml_for_active_property()
        active_collection = rprop.PropertyCollection()
        active_collection.set_support(support = grid)
        active_collection.create_xml(None,
                                     None,
                                     'ACTIVE',
                                     'active',
                                     p_uuid = grid.active_property_uuid,
                                     discrete = True,
                                     add_min_max = False,
                                     find_local_property_kinds = True)

    return ijk


def __add_as_part(add_relationships, ext_uuid, grid, ijk, write_geometry):
    grid.model.add_part('obj_IjkGridRepresentation', grid.uuid, ijk)
    if add_relationships:
        if grid.stratigraphic_column_rank_uuid is not None and grid.stratigraphic_units is not None:
            grid.model.create_reciprocal_relationship(ijk, 'destinationObject',
                                                      grid.model.root_for_uuid(grid.stratigraphic_column_rank_uuid),
                                                      'sourceObject')
        if write_geometry:
            # create 2 way relationship between IjkGrid and Crs
            if grid.crs is None:
                crs_root = grid.model.root_for_uuid(grid.crs_uuid)
            else:
                crs_root = grid.crs.root
            grid.model.create_reciprocal_relationship(ijk, 'destinationObject', crs_root, 'sourceObject')
            # create 2 way relationship between IjkGrid and Ext
            ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
            ext_node = grid.model.root_for_part(ext_part)
            grid.model.create_reciprocal_relationship(ijk, 'mlToExternalPartProxy', ext_node, 'externalPartProxyToMl')
        # create relationship with parent grid
        if grid.parent_window is not None and grid.parent_grid_uuid is not None:
            grid.model.create_reciprocal_relationship(ijk, 'destinationObject',
                                                      grid.model.root_for_uuid(grid.parent_grid_uuid), 'sourceObject')


def __add_geometry_xml(ext_uuid, grid, ijk):
    geom = rqet.SubElement(ijk, ns['resqml2'] + 'Geometry')
    geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'IjkGridGeometry')
    geom.text = '\n'
    # the remainder of this function is populating the geometry node
    grid.model.create_crs_reference(crs_uuid = grid.crs_uuid, root = geom)
    k_dir = rqet.SubElement(geom, ns['resqml2'] + 'KDirection')
    k_dir.set(ns['xsi'] + 'type', ns['resqml2'] + 'KDirection')
    if grid.k_direction_is_down:
        k_dir.text = 'down'
    else:
        k_dir.text = 'up'
    handed = rqet.SubElement(geom, ns['resqml2'] + 'GridIsRighthanded')
    handed.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
    handed.text = str(grid.grid_is_right_handed).lower()
    p_shape = rqet.SubElement(geom, ns['resqml2'] + 'PillarShape')
    p_shape.set(ns['xsi'] + 'type', ns['resqml2'] + 'PillarShape')
    p_shape.text = grid.pillar_shape
    points_node = rqet.SubElement(geom, ns['resqml2'] + 'Points')
    points_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
    points_node.text = '\n'
    coords = rqet.SubElement(points_node, ns['resqml2'] + 'Coordinates')
    coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
    coords.text = '\n'
    grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'Points', root = coords)
    if always_write_pillar_geometry_is_defined_array or not grid.geometry_defined_for_all_pillars(cache_array = True):

        pillar_def = rqet.SubElement(geom, ns['resqml2'] + 'PillarGeometryIsDefined')
        pillar_def.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanHdf5Array')
        pillar_def.text = '\n'

        pd_values = rqet.SubElement(pillar_def, ns['resqml2'] + 'Values')
        pd_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        pd_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'PillarGeometryIsDefined', root = pd_values)

    else:

        pillar_def = rqet.SubElement(geom, ns['resqml2'] + 'PillarGeometryIsDefined')
        pillar_def.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanConstantArray')
        pillar_def.text = '\n'

        pd_value = rqet.SubElement(pillar_def, ns['resqml2'] + 'Value')
        pd_value.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        pd_value.text = 'true'

        pd_count = rqet.SubElement(pillar_def, ns['resqml2'] + 'Count')
        pd_count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        pd_count.text = str((grid.extent_kji[1] + 1) * (grid.extent_kji[2] + 1))
    if always_write_cell_geometry_is_defined_array or not grid.geometry_defined_for_all_cells(cache_array = True):

        cell_def = rqet.SubElement(geom, ns['resqml2'] + 'CellGeometryIsDefined')
        cell_def.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanHdf5Array')
        cell_def.text = '\n'

        cd_values = rqet.SubElement(cell_def, ns['resqml2'] + 'Values')
        cd_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        cd_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'CellGeometryIsDefined', root = cd_values)

    else:

        cell_def = rqet.SubElement(geom, ns['resqml2'] + 'CellGeometryIsDefined')
        cell_def.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanConstantArray')
        cell_def.text = '\n'

        cd_value = rqet.SubElement(cell_def, ns['resqml2'] + 'Value')
        cd_value.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        cd_value.text = 'true'

        cd_count = rqet.SubElement(cell_def, ns['resqml2'] + 'Count')
        cd_count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        cd_count.text = str(grid.nk * grid.nj * grid.ni)
    if grid.has_split_coordinate_lines:
        scl = rqet.SubElement(geom, ns['resqml2'] + 'SplitCoordinateLines')
        scl.set(ns['xsi'] + 'type', ns['resqml2'] + 'ColumnLayerSplitCoordinateLines')
        scl.text = '\n'

        scl_count = rqet.SubElement(scl, ns['resqml2'] + 'Count')
        scl_count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        scl_count.text = str(grid.split_pillars_count)

        pi_node = rqet.SubElement(scl, ns['resqml2'] + 'PillarIndices')
        pi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        pi_node.text = '\n'

        pi_null = rqet.SubElement(pi_node, ns['resqml2'] + 'NullValue')
        pi_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        pi_null.text = str((grid.extent_kji[1] + 1) * (grid.extent_kji[2] + 1))

        pi_values = rqet.SubElement(pi_node, ns['resqml2'] + 'Values')
        pi_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        pi_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid, grid.uuid, 'PillarIndices', root = pi_values)

        cpscl = rqet.SubElement(scl, ns['resqml2'] + 'ColumnsPerSplitCoordinateLine')
        cpscl.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlJaggedArray')
        cpscl.text = '\n'

        elements = rqet.SubElement(cpscl, ns['resqml2'] + 'Elements')
        elements.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        elements.text = '\n'

        el_null = rqet.SubElement(elements, ns['resqml2'] + 'NullValue')
        el_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        el_null.text = str(grid.extent_kji[1] * grid.extent_kji[2])

        el_values = rqet.SubElement(elements, ns['resqml2'] + 'Values')
        el_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        el_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid,
                                           grid.uuid,
                                           'ColumnsPerSplitCoordinateLine/elements',
                                           root = el_values)

        c_length = rqet.SubElement(cpscl, ns['resqml2'] + 'CumulativeLength')
        c_length.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        c_length.text = '\n'

        cl_null = rqet.SubElement(c_length, ns['resqml2'] + 'NullValue')
        cl_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        cl_null.text = '0'

        cl_values = rqet.SubElement(c_length, ns['resqml2'] + 'Values')
        cl_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        cl_values.text = '\n'

        grid.model.create_hdf5_dataset_ref(ext_uuid,
                                           grid.uuid,
                                           'ColumnsPerSplitCoordinateLine/cumulativeLength',
                                           root = cl_values)
    if grid.time_index is not None:
        assert grid.time_series_uuid is not None

        ti_node = rqet.SubElement(geom, ns['resqml2'] + 'TimeIndex')
        ti_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TimeIndex')
        ti_node.text = '\n'

        index_node = rqet.SubElement(ti_node, ns['resqml2'] + 'Index')
        index_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        index_node.text = str(grid.time_index)

        grid.model.create_ref_node('TimeSeries',
                                   grid.model.title(uuid = grid.time_series_uuid),
                                   grid.time_series_uuid,
                                   content_type = 'obj_TimeSeries',
                                   root = ti_node)



"""_stratigraphic_column.py: RESQML StratigraphicColumn class."""

version = '24th November 2021'

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
from resqpy.olio.base import BaseResqpy
from resqpy.strata._strata_common import _index_attr
from resqpy.strata._stratigraphic_column_rank import StratigraphicColumnRank


class StratigraphicColumn(BaseResqpy):
    """Class for RESQML stratigraphic column objects.

    RESQML documentation:

       A global interpretation of the stratigraphy, which can be made up of
       several ranks of stratigraphic unit interpretations.

       All stratigraphic column rank interpretations that make up a stratigraphic column
       must be ordered by age.
    """

    resqml_type = 'StratigraphicColumn'

    # todo: integrate with EarthModelInterpretation class which can optionally refer to a stratigraphic column

    def __init__(self, parent_model, uuid = None, rank_uuid_list = None, title = None, extra_metadata = None):
        """Initialises a Stratigraphic Column object.

        arguments:
           parent_model (model.Model): the model with which the new stratigraphic column will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML stratigraphic column from which
              this object will be initialised
           rank_uuid_list (list of uuid, optional): if not initialising for an existing stratigraphic column,
              the ranks can be established from this list of uuids for existing stratigraphic column ranks;
              ranks should be ordered from geologically oldest to youngest with increasing index values
           title (str, optional): the citation title of the new stratigraphic column; ignored if uuid is not None
           extra_metadata (dict, optional): extra metadata items for the new feature
        """

        self.ranks = []  # list of Stratigraphic Column Rank Interpretation objects, maintained in rank index order

        super().__init__(model = parent_model, uuid = uuid, title = title, extra_metadata = extra_metadata)

        if self.root is None and rank_uuid_list:
            for rank_uuid in rank_uuid_list:
                rank = StratigraphicColumnRank(self.model, uuid = rank_uuid)
                self.add_rank(rank)

    def _load_from_xml(self):
        """Loads class specific attributes from xml for an existing RESQML object; called from BaseResqpy."""
        rank_node_list = rqet.list_of_tag(self.root, 'Ranks')
        assert rank_node_list is not None, 'no stratigraphic column ranks in xml for stratigraphic column'
        for rank_node in rank_node_list:
            rank = StratigraphicColumnRank(self.model, uuid = rqet.find_tag_text(rank_node, 'UUID'))
            self.add_rank(rank)

    def iter_ranks(self):
        """Yields the stratigraphic column ranks which constitute this stratigraphic colunn."""

        for rank in self.ranks:
            yield rank

    def add_rank(self, rank):
        """Adds another stratigraphic column rank to this stratigraphic column.

        arguments:
           rank (StratigraphicColumnRank): an established rank to be added to this stratigraphic column

        note:
           ranks should be ordered from geologically oldest to youngest, with increasing index values
        """

        assert rank is not None and rank.index is not None
        self.ranks.append(rank)
        self._sort_ranks()

    def _sort_ranks(self):
        """Sort the list of ranks, in situ, by their index values."""
        self.ranks.sort(key = _index_attr)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False.

        arguments:
           other (StratigraphicColumn): the other stratigraphic column to compare this one against
           check_extra_metadata (bool, default True): if True, then extra metadata items must match for the two
              columns to be deemed equivalent; if False, extra metadata is ignored in the comparison

        returns:
           bool: True if this stratigraphic column is essentially the same as the other; False otherwise
        """

        if not isinstance(other, StratigraphicColumn):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if len(self.ranks) != len(other.ranks):
            return False
        for rank_a, rank_b in zip(self.ranks, other.ranks):
            if rank_a != rank_b:
                return False
        if check_extra_metadata and not rqo.equivalent_extra_metadata(self, other):
            return False
        return True

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates xml tree for a stratigraphic column object.

        arguments:
           add_as_part (bool, default True): if True, the stratigraphic column is added to the parent model as a high level part
           add_relationships (bool, default True): if True and add_as_part is True, a relationship is created with each of
              the referenced stratigraphic column rank interpretations
           originator (str, optional): if present, is used as the originator field of the citation block
           reuse (bool, default True): if True, the parent model is inspected for any equivalent stratigraphic column and,
              if found, the uuid of this stratigraphic column is set to that of the equivalent part

        returns:
           lxml.etree._Element: the root node of the newly created xml tree for the stratigraphic column
        """

        assert self.ranks, 'attempt to create xml for stratigraphic column without any contributing ranks'

        if reuse and self.try_reuse():
            return self.root

        sci = super().create_xml(add_as_part = False, originator = originator)

        assert sci is not None

        for rank in self.ranks:
            self.model.create_ref_node('Ranks',
                                       rank.title,
                                       rank.uuid,
                                       content_type = 'obj_StratigraphicColumnRankInterpretation',
                                       root = sci)

        if add_as_part:
            self.model.add_part('obj_StratigraphicColumn', self.uuid, sci)
            if add_relationships:
                for rank in self.ranks:
                    self.model.create_reciprocal_relationship(sci, 'destinationObject', rank.root, 'sourceObject')

        return sci



"""tetra_grid.py: resqpy TetraGrid class module."""

version = '24th November 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.grid as grr
import resqpy.olio.volume as vol
from resqpy.unstructured._unstructured_grid import UnstructuredGrid, valid_cell_shapes


class TetraGrid(UnstructuredGrid):
    """Class for unstructured grids where every cell is a tetrahedron."""

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 cache_geometry = False,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a new resqpy TetraGrid object (RESQML UnstructuredGrid with cell shape tetrahedral)

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created TetraGrid object
        """

        super().__init__(parent_model = parent_model,
                         uuid = uuid,
                         find_properties = find_properties,
                         geometry_required = True,
                         cache_geometry = cache_geometry,
                         cell_shape = 'tetrahedral',
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            assert grr.grid_flavour(self.root) == 'TetraGrid'
            self.check_tetra()

        self.grid_representation = 'TetraGrid'  #: flavour of grid; not much used

    def check_tetra(self):
        """Checks that each cell has 4 faces and each face has 3 nodes."""

        assert self.cell_shape == 'tetrahedral'
        self.cache_all_geometry_arrays()
        assert self.faces_per_cell_cl is not None and self.nodes_per_face_cl is not None
        assert self.faces_per_cell_cl[0] == 4 and np.all(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1] == 4)
        assert self.nodes_per_face_cl[0] == 3 and np.all(self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1] == 3)

    def face_centre_point(self, face_index):
        """Returns a nominal centre point for a single face calculated as the mean position of its nodes.

        note:
           this is a nominal centre point for a face and not generally its barycentre
        """

        self.cache_all_geometry_arrays()
        start = 0 if face_index == 0 else self.nodes_per_face_cl[face_index - 1]
        return np.mean(self.points_cached[self.nodes_per_face[start:start + 3]], axis = 0)

    def volume(self, cell):
        """Returns the volume of a single cell.

        arguments:
           cell (int): the index of the cell for which the volume is required

        returns:
           float being the volume of the tetrahedral cell; units of measure is implied by crs units
        """

        self.cache_all_geometry_arrays()
        abcd = self.points_cached[self.distinct_node_indices_for_cell(cell)]
        assert abcd.shape == (4, 3)
        return vol.tetrahedron_volume(abcd[0], abcd[1], abcd[2], abcd[3])

    def grid_volume(self):
        """Returns the sum of the volumes of all the cells in the grid.

        returns:
           float being the total volume of the grid; units of measure is implied by crs units
        """

        v = 0.0
        for cell in range(self.cell_count):
            v += self.volume(cell)
        return v

    @classmethod
    def from_unstructured_cell(cls, u_grid, cell, title = None, extra_metadata = {}, set_handedness = False):
        """Instantiates a small TetraGrid representing a single cell from an UnstructuredGrid as a set of tetrahedra."""

        def _min_max(a, b):
            if a < b:
                return (a, b)
            else:
                return (b, a)

        if not title:
            title = str(u_grid.title) + f'_cell_{cell}'

        assert u_grid.cell_shape in valid_cell_shapes
        u_grid.cache_all_geometry_arrays()
        u_cell_faces = u_grid.face_indices_for_cell(cell)
        u_cell_nodes = u_grid.distinct_node_indices_for_cell(cell)

        # create an empty TetreGrid
        tetra = cls(u_grid.model, title = title, extra_metadata = extra_metadata)
        tetra.crs_uuid = u_grid.crs_uuid

        u_cell_node_count = len(u_cell_nodes)
        assert u_cell_node_count >= 4
        u_cell_face_count = len(u_cell_faces)
        assert u_cell_face_count >= 4

        # build attributes, depending on the shape of the individual unstructured cell

        if u_cell_node_count == 4:  # cell is tetrahedral

            assert u_cell_face_count == 4
            tetra.set_cell_count(1)
            tetra.face_count = 4
            tetra.faces_per_cell_cl = np.array((4,), dtype = int)
            tetra.faces_per_cell = np.arange(4, dtype = int)
            tetra.node_count = 4
            tetra.nodes_per_face_cl = np.arange(3, 3 * 4 + 1, 3, dtype = int)
            tetra.nodes_per_face = np.array((0, 1, 2, 0, 3, 1, 1, 3, 2, 2, 3, 0), dtype = int)
            tetra.cell_face_is_right_handed = np.ones(4, dtype = bool)
            tetra.points_cached = u_grid.points_cached[u_cell_nodes].copy()

        # todo: add optimised code for pyramidal (and hexahedral?) cells

        else:  # generic case: add a node at centre of unstructured cell and divide faces into triangles

            tetra.node_count = u_cell_node_count + 1
            tetra.points_cached = np.empty((tetra.node_count, 3))
            tetra.points_cached[:-1] = u_grid.points_cached[u_cell_nodes].copy()
            tetra.points_cached[-1] = u_grid.centre_point(cell = cell)
            centre_node = tetra.node_count - 1

            u_cell_nodes = list(u_cell_nodes)  # to allow simple index usage below

            # build list of distinct edges used by cell
            u_cell_edge_list = u_grid.distinct_edges_for_cell(cell)
            u_edge_count = len(u_cell_edge_list)
            assert u_edge_count >= 4

            t_cell_list = []  # list of 4-tuples of ints, being local face indices for tetra cells

            # create an internal tetra face for each edge, using centre point as third node
            # note: u_a, u_b are a sorted pair, and u_cell_nodes is also aorted, so t_a, t_b are a sorted pair
            t_face_list = []  # list of triple ints each triplet being local node indices for a triangular face
            for u_a, u_b in u_cell_edge_list:
                t_a, t_b = u_cell_nodes.index(u_a), u_cell_nodes.index(u_b)
                t_face_list.append((t_a, t_b, centre_node))

            # for each unstructured face, create a Delauney triangulation; create a tetra face for each
            # triangle in the triangulation; create internal tetra faces for each of the internal edges in
            # the triangulation; and create a tetra cell for each triangle in the triangulation
            # note: the resqpy Delauney triangulation is for a 2D system, so here the unstructured face
            # is projected onto a planar approximation defined by the face centre point and an average
            # normal vector
            for fi in u_cell_faces:
                triangulated_face = u_grid.face_triangulation(fi)
                for u_a, u_b, u_c in triangulated_face:
                    t_a, t_b, t_c = u_cell_nodes.index(u_a), u_cell_nodes.index(u_b), u_cell_nodes.index(u_c)
                    t_cell_faces = [len(t_face_list)]  # local face index for this triangle
                    t_face_list.append((t_a, t_b, t_c))
                    tri_edges = np.array([_min_max(t_a, t_b), _min_max(t_b, t_c), _min_max(t_c, t_a)], dtype = int)
                    for e_a, e_b in tri_edges:
                        try:
                            pos = t_face_list.index((e_a, e_b, centre_node))
                        except ValueError:
                            pos = len(t_face_list)
                            t_face_list.append((e_a, e_b, centre_node))
                        t_cell_faces.append(pos)
                    t_cell_list.append(tuple(t_cell_faces))

            # everything is now ready to populate the tetra grid attributes (apart from handedness)
            tetra.set_cell_count(len(t_cell_list))
            tetra.face_count = len(t_face_list)
            tetra.faces_per_cell_cl = np.arange(4, 4 * tetra.cell_count + 1, 4, dtype = int)
            tetra.faces_per_cell = np.array(t_cell_list, dtype = int).flatten()
            tetra.nodes_per_face_cl = np.arange(3, 3 * tetra.face_count + 1, 3, dtype = int)
            tetra.nodes_per_face = np.array(t_face_list, dtype = int).flatten()

            tetra.cell_face_is_right_handed = np.ones(len(tetra.faces_per_cell), dtype = bool)
            if set_handedness:
                # TODO: set handedness correctly and make default for set_handedness True
                raise NotImplementedError('code not written to set handedness for tetra from unstructured cell')

        tetra.check_tetra()

        return tetra

    # todo: add tetra specific method for centre_point()
